#
# With the expansion of the number of command line tools built within the wf/commands.py file, it's getting a bit
# difficult to find what you need.  This file contains the various CLI interfaces to the ingestion-focused workflows.
#
import argparse as ap
import os
import sys
import time
import logging


from pyat.wf import wf, WorkflowTracker
from pathlib import Path
from pyat import version
from pycapo import CapoConfig
from pyat.wf.ous_wf_interfaces import lookup_single_locator, lookup_project_code, lookup_eb_uid

# Ingestion Workflow CLIs
def _make_ingestion_parser(ingestion_description, major_arg):
    r"""Shared Command-line Ingestion Launch Parser"""
    result = ap.ArgumentParser(description=ingestion_description.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. nmtest, mnprod')
    result.add_argument('-p', '--path', action='store', default='',
                        help='Path to data (overrides CAPO setting)')
    result.add_argument('fileInfo', type = str, action='store', nargs='+', help=major_arg)
    return result


def launch_sdm_ingestion():
    r"""CLI for SDM Ingestion Workflow"""
    sdm_description = """Manual SDM Ingestion, version {}: Initiates an SDM Ingestion workflow for each listed execution
    block."""
    sdm_argument = 'FileSet Identifiers(s) to ingest'
    parser = _make_ingestion_parser(sdm_description, sdm_argument)
    args = parser.parse_args()

    # Automatic Profile Determination:
    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    #
    # Command line we want is:
    #
    # wf -P [test|nmprod] -p workflowName=SdmIngestion -p dataSetId=arg runSdmIngestionWorkflow
    #   or
    # wf -P [nmtest|nmprod] -p workflowName=SdmIngestion -p SDMPath=arg -p datasetId=arg runSdmIngestionWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=SdmIngestion -p processingSite=DSOC -P " #profile here
    clArgPiece = " -p datasetId=" #fileSetId here
    if '' != args.path:
        clPathPiece = " -p SDMPath=" + args.path
        # insert the casa parameter here, since it is fixed for all fileSetIds
        clArgPiece = clPathPiece + clArgPiece

    clFinish = " runSdmIngestionWorkflow"


    #
    # Loop over the provided fileSetIds and send a command for each:
    #
    for id in args.fileInfo:
        wfArgs = clBase + os.environ['CAPO_PROFILE'] + clArgPiece + id + clFinish
        #print("wf "+wfArgs)
        wf(args=wfArgs.split(' '))
        time.sleep(0.25) # 1/4 second delay between workflow launches for saftey


def launch_bdf_ingestion():
    r"""CLI for BDF Ingestion Workflow"""
    bdf_description = """Manual BDF Ingestion, version {}: Launches a BDF ingestion workflow for the listed files."""
    bdf_argument = 'BDF File(s) to Ingest'
    parser = _make_ingestion_parser(bdf_description, bdf_argument)
    args = parser.parse_args()

    # Automatic Profile Determination:
    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    #
    # Command line we want is:
    #
    # wf -P [test|nmprod] -p workflowName=SdmIngestion -p bdfs=[list from arg] runBdfIngestionWorkflow
    #   or
    # wf -P [nmtest|nmprod] -p workflowName=SdmIngestion -p SDMPath=arg -p bdfs=[list from arg] runBdfIngestionWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=BdfIngestion -p processingSite=DSOC -P " #profile here
    clArgPiece = " -p bdfs=" #fileSetId here
    if '' != args.path:
        clPathPiece = " -p BDFPath=" + args.path
        # insert the casa parameter here, since it is fixed for all fileSetIds
        clArgPiece = clPathPiece + clArgPiece

    clFinish = " runBdfIngestionWorkflow"

    # We need to massage the list of filenames into a string without spaces
    temp_string = str(args.fileInfo)
    bdf_string = temp_string.replace(' ', '')

    # Put it all together (all files in one command:
    wf_arguments = clBase + os.environ['CAPO_PROFILE'] + clArgPiece + bdf_string + clFinish
    #print("wf "+wf_arguments)
    wf(args=wf_arguments.split(' '))


def launch_idifits_ingestion():
    r"""CLI for VLBA IDI Fits file ingestion"""
    idi_description = """Manual VLBA IDI FITS Ingestion, version {}: Initiates an ingestion workflow for each idi fits 
    file in the named directory."""
    idi_argument = 'Folder named for the project & segment to process (all contained IDIFITS files will be ingested)'
    parser = _make_ingestion_parser(idi_description, idi_argument)
    args = parser.parse_args()

    # Automatic Profile Determination:
    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    # We need our full path, so check for the -p override, or extract it from CAPO
    settings = CapoConfig(profile=os.environ['CAPO_PROFILE'])
    path_base_string = settings.getstring('archive-ingestion.VLBAPath')
    if '' != args.path:
        path_base_string = args.path

    base_ingestion = Path(path_base_string)

    # A couple checks:
    if not base_ingestion.exists():
        logging.error(f"Base ingestion path {path_base_string} doesn't exist!")
        exit(-1)
    if not base_ingestion.is_dir() or not base_ingestion.is_absolute():
        logging.error(f"Base ingestion path is inappropriate: {path_base_string}")
        exit(-1)


    #
    # So, for each directory we're given, we need to:
    #
    # 1) extract the legacy ID to provide the workflow
    # 2) obtain a listing of the fits files, and launch a workflow to ingest each one.
    #
    for directory_name in args.fileInfo:
        # The directories are named in a pattern, which makes my life easier:
        # (ProjectCode)(Segment).(CorrelatorPass)
        #   The project code (what we call LegacyId) follows AA111
        #   The segment follows A[0-9]*
        #directory_components = directory_name.split(".")
        # legacy id is the first 5 characters of the first piece, which is guaranteed to exist from split()
        #legacy_id = directory_components[0][:5]
        # print(f"The Legacy id for {directory_name} is {legacy_id}.")

        #
        # Now get the filenames:
        #
        ingestion_path = base_ingestion / directory_name

        files = ingestion_path.glob("*.idifits")
        for file in files:
            print(f"\tLaunching workflow to ingest: {file.name} from {str(ingestion_path)}")

            #
            # The files have names in the format:
            #       VLBA_DQ840_dq840_BIN0_SRC0_02_181016T191428.idifits*
            #
            #       VLBA_[PC][Segment]_.......
            #   We can use this to extract the project code on a per-file basis.
            #   This is safer for a bulk-ingestion of legacy data.
            name_components = file.name.split("_")

            if 3 >= len(name_components):
                # if we don't have the requisite number of pieces, then skip it!
                print(f"Error!  File {file.name} has only {len(name_components)} subsections.")
                continue

            legacy_id = name_components[1][:5]
            #print(f"Legacy id for {file.name} is {legacy_id}.")

            #
            # Command line we want to emulate:
            #
            # wf -P nmtest
            #   -p ingestionPath=/home/e2earchive/hold/difx/BD215G3.bd215g3.cata
            #   -p subDirOrFileName=VLBA_BD215G3_bd215g3_BIN0_SRC1_0_181001T142116.idifitsVLBA_BD215G3_bd215g3_BIN0_SRC1_0_181001T142116.idifits
            #   -p projectCode=BD215 -p ingestionType=vlba -p workflowName=VlbaIngestion runVlbaIngestionWorkflow
            #
            # The pieces of the string
            cl_base       = "-p workflowName=VlbaIngestion -p ingestionType=vlba -s DSOC -P " #profile here
            cl_arg_piece  = " -p subDirOrFileName=" + file.name
            cl_path_piece = " -p ingestionPath=" + str(ingestion_path)
            cl_code_piece = " -p projectCode=" + legacy_id
            cl_finish     = " runVlbaIngestionWorkflow"

            total_command = cl_base + os.environ['CAPO_PROFILE'] + cl_arg_piece + cl_path_piece + cl_code_piece + cl_finish
            # print(f"\tFor WF: {total_command}")

            wf(args=total_command.split(' '))
            time.sleep(1.25)  # Wait a bit between workflow launches

        print(f"Finished with {directory_name}.\n")


#
# Image ingestion is a slightly different beast:  You point to a directory, and we collect the FITS files
# and bundle the ancillary files into a tar bundle for ingestion.
#
# For image reingestion, you assume the directory already has just the FITS files and the tar bundle.
#
def _make_image_ingestion_parser(ingestion_description):
    r"""Shared Command-line Ingestion Launch Parser"""
    result = ap.ArgumentParser(description=ingestion_description.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. nmtest, mnprod')
    result.add_argument('projectCode', action='store', type=str, help='Project for these Images')
    result.add_argument('imageSetLoc', action='store', type=str, nargs='+',
                        help='Directory from which to draw FITS images & ancillary file(s)')
    return result


def launch_image_ingestion():
    r"""CLI for the Image Ingestion Workflow"""
    image_description = """Manual Image Ingestion, version {}: Initiates an image ingestion workflow 
                         for each listed location (note, all sets ingested with a single command will 
                         share the same project)"""
    parser = _make_image_ingestion_parser(image_description)
    args = parser.parse_args()

    # Automatic Profile Determination:
    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    #
    # Command line we want is:
    #
    # wf -P [test|nmprod] -p workflowName=SdmIngestion -p dataSetId=arg runSdmIngestionWorkflow
    #   or
    # wf -P [nmtest|nmprod] -p workflowName=SdmIngestion -p SDMPath=arg -p datasetId=arg runSdmIngestionWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=QuicklookImageIngestion -p processingSite=DSOC -p ingestionType=vlass_quicklook -P " #profile here
    clProjectPiece = " -p projectCode="+args.projectCode
    clArgPath = " -p ingestionPath="
    clArgPiece = " -p subDirOrFileName="
    clFinish = " runQuicklookIngestionWorkflow"


    #
    # Loop over the provided directories and send a command for each:
    #
    for id in args.imageSetLoc:
        id_as_path = Path(id)
        subdirectory = str(id_as_path.name)
        ingest_path  = str(id_as_path.parent)
        wfArgs = clBase + os.environ['CAPO_PROFILE'] + clProjectPiece + clArgPath + ingest_path + clArgPiece + subdirectory + clFinish
        # print("wf "+wfArgs)
        wf(args=wfArgs.split(' '))
        time.sleep(0.25) # 1/4 second delay between workflow launches for saftey


def launch_image_reingestion():
    r"""CLI for the Image Ingestion Workflow"""
    image_description = """Manual Image Metadat Reingestion, version {}: provided the location of an image set
    (Fits files + tar) this workflow deletes the existing metadata & rebuilds the entries.  A full cycle
    reingestion workflow is possible, but not yet necessary. """
    parser = _make_image_ingestion_parser(image_description)
    args = parser.parse_args()

    # Automatic Profile Determination:
    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    #
    # Command line we want is:
    #
    # wf -P [test|nmprod] -p workflowName=SdmIngestion -p dataSetId=arg runSdmIngestionWorkflow
    #   or
    # wf -P [nmtest|nmprod] -p workflowName=SdmIngestion -p SDMPath=arg -p datasetId=arg runSdmIngestionWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=ImageReingestion -p processingSite=DSOC -P " #profile here
    clProjectPiece = " -p projectCode="+args.projectCode
    clArgPiece = " -p sourceDirectory="
    clFinish = " runImageReingestionWorkflow"


    #
    # Loop over the provided directories and send a command for each:
    #
    for id in args.imageSetLoc:
        wfArgs = clBase + os.environ['CAPO_PROFILE'] + clProjectPiece + clArgPiece + id + clFinish
        # print("wf "+wfArgs)
        wf(args=wfArgs.split(' '))
        time.sleep(0.25) # 1/4 second delay between workflow launches for saftey


_WORKFLOW_NAME = 'AoiIngestionWorkflow'

_RUN_AOI_DESCRIPTION = """AAT/PPI ALMA Optimized Image Ingestion, version {}: This tool initiates AOI for the supplied 
    fileSetId(s). Upon successful processing, the request state is changed to PENDINGQA, and the user is notified
    that the request awaits QA. When the data have passed QA, the data are delivered, request state is changed to 
    COMPLETE and the user sent a download link. """


def find_mous_id(path_to_data):
    r""" Somewhere in the path will be the external name,
         either just below the top-level directory (if no request ID)
         or underneath top-level-dir/request_id/.
    """

    external_name = None
    filename      = None
    rawdata_dir   = None
    for base, dirs, files in os.walk(path_to_data):
        for dir in dirs:
            # A directory name corresponding to the mousId will have underscores rather than
            # colon & slashes
            if str(dir).endswith('rawdata'):
                rawdata_dir = dir
                break

            if rawdata_dir is not None:
                for base, dirs, files in os.walk(path_to_data):
                    # the ASA_OUS_ID will be hidden in the name of a data file;
                    # it -won't- be a directory name
                    for file in files:
                        if str(file).startswith("member."):
                            filename = str(file)
                            break

            if filename is not None:
                parts = filename.split('.')
                external_name = parts[1]
                break

        if external_name is not None:
            break

    if external_name is not None:
        # translate to OUS ID format
        mousId = external_name.replace("___", "://").replace("_", "/")
        return mousId

    return None


def find_working_subdirectory(request_path):
    """
        A utility for handling external requests.  The tool is
        provided a directory named for the request id, and we
        need to provide the name of the actual working directory
        underneath.

        Note: for an externally-initiated AUDI there should be
        a single subdirectory, if we find more than one, we've
        got an issue.
    :return: name of subdirectory (or None if an error occurs)
    """

    request_subdir = None

    for component in request_path.iterdir():
        if component.is_dir():
            # we really only ought to have one...
            if request_subdir is None:
                request_subdir = component.name
            else:
                # if we've already filled in an answer
                # and found another subdirectory....
                # something's wrong.
                return None

    return request_subdir


def launch_alma_image_ingest_qa():
    r"""Launch runAoiIngestionWorkflow for data in given image-qa subdirectory
        Based on ingest_wf_interfaces.launch_image_ingestion() -- thanks, Jim
    """

    # example command lines:
    #
    #   audiPass uid___A002_Xd3607d_X2b0d_2018_10_16_T14_19_15.962 -E jgoldste@nrao.edu
    #   audiPass 299837780
    #
    # subdirectory is assumed to exist under AlmaImagingWorkflowSettings.qaDirectory
    #

    # parse our arguments; get profile and other info we need
    args = _make_audi_pass_parser().parse_args()
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        capo_profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = capo_profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(capo_profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    capo_profile = os.environ['CAPO_PROFILE']
    configuration = CapoConfig(profile=capo_profile)
    top_level_dir = configuration.get('edu.nrao.archive.workflow.config.AlmaImagingWorkflowSettings.qaDirectory')

    # stray '/' cause problems:
    provided_qa_dir = args.sourceDirectory.replace('/', '')
    path_to_data = os.path.join(top_level_dir, provided_qa_dir)

    if not os.path.isdir(path_to_data):
        raise NotADirectoryError(f'{path_to_data} not found')

    # Check if we were given a request Id
    try:
        request_id = int(provided_qa_dir)
        # with a request id, our subdirectory of interest is one level down,
        # go find it
        subdirectory = find_working_subdirectory(Path(top_level_dir, provided_qa_dir))

        if subdirectory is None:
            print(f'Error: Could not identify the working directory under {provided_qa_dir}.')

            exit(-1)

    except ValueError:
        # If this fails, then we have an internal run
        # no request id, just directly given the working directory
        request_id = None
        subdirectory = provided_qa_dir

    mous_id = find_mous_id(path_to_data)

    if mous_id is not None:
        try:
            project_code = lookup_project_code(mous_id)
        except TypeError:
            print(f'>>> No project code found for MOUS ID {mous_id}')
            exit(-1)
    else:
        print(f'Error: MOUS ID not found in {provided_qa_dir}.')
        exit(-1)

    if project_code is None:
        print(f'Error: No project code found for MOUS ID {mous_id}.')
        exit(-1)

    my_tracker = WorkflowTracker(capo_profile)
    monitor_key = my_tracker.prepareWorkflowTracking(_WORKFLOW_NAME, email=args.email)

    launch_audi_pass(mous_id, project_code, subdirectory, request_id, monitor_key)

    # Now wait for information:
    my_tracker.waitForReply()

    print(f'Image ingestion QA ({my_tracker.getLoggingId()}) running in: {my_tracker.getWorkingDirectory()}')


def _make_audi_pass_parser():

    r"""Command-line AOI Parser"""

    result = ap.ArgumentParser(description=_RUN_AOI_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)

    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. nmtest, mnprod')

    result.add_argument('sourceDirectory', action='store', type=str,
                        help='image-qa subdirectory containing data')

    result.add_argument('-E', '--email', action='store', default=None,
                        help='Optional email address for a completion message')

    # Placeholder; workflow can't do this yet.
    # result.add_argument('-D', '--destination', action='store', default='', help='Override location for AUDI pass')

    return result


def launch_audi_pass(data_set_id, project_code, subdir, request_id, tracking_key):
    # build pieces of command line

    capo_profile =  os.environ['CAPO_PROFILE']
    arguments = '-P ' + capo_profile
    arguments += ' -s NAASC'
    arguments += ' -p telescope=ALMA'
    arguments += ' -p projectCode=' + project_code
    arguments += ' -p workflowName=AoiIngestionWorkflow'
    arguments += ' -p subDirOrFileName=' +subdir
    arguments += ' -p dataSetId=' + data_set_id
    if request_id is not None:
        arguments += ' -p requestId=' + str(request_id)
    arguments += ' -p ingestionType=alma_aoi'
    arguments += ' -p cliCorrId=' + tracking_key
    arguments += ' run'+ _WORKFLOW_NAME

    wf(args=arguments.split(' '))

#
#
#  Collection-specific Variants
#
#

def _make_realfast_parser(ingestion_description, major_arg):
    r"""Shared Command-line Ingestion Launch Parser"""
    result = ap.ArgumentParser(description=ingestion_description.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. nmtest, mnprod')
    result.add_argument('-s', '--sdm_path', action='store', default='',
                        help='Path to the RealFast SDM (overrides CAPO setting)')
    result.add_argument('-p', '--png_path', action='store', default='',
                        help='Path to the candidate PNG files (overrides CAPO setting)')
    result.add_argument('sdmName', type = str, action='store', nargs='+', help=major_arg)
    return result


def launch_realfast_sdm_ingestion():
    r"""CLI for RealFast SDM Ingestion Workflow"""
    rf_description = """RealFast SDM Ingestion, version {}: Initiates an ingestion workflow for the SDM and ancillary 
    files for each execution block listed."""
    rf_argument = 'FileSet Identifiers(s) to ingest'
    parser = _make_realfast_parser(rf_description, rf_argument)
    args = parser.parse_args()

    # Automatic Profile Determination:
    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == args.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != args.profile:
        os.environ['CAPO_PROFILE'] = args.profile

    #
    # Very similar to regular SDM ingestion, the command line we want is:
    #
    # wf -P [test|nmprod] -p workflowName=RfSdmIngestion -p dataSetId=arg runRealFastSdmIngestionWorkflow
    #   or
    # wf -P [nmtest|nmprod] -p workflowName=RfSdmIngestion -p SDMPath=arg -p pngPath=arg -p datasetId=arg runRealFastSdmIngestionWorkflow
    #
    # The pieces of the string
    clBase = "-p workflowName=RfSdmIngestion -p processingSite=DSOC -P" #profile here
    clArgPiece = " -p ingestionType=realfast_sdm -p datasetId=" #fileSetId here
    if '' != args.sdm_path:
        clPathPiece = " -p SDMPath=" + args.sdm_path
        # insert the casa parameter here, since it is fixed for all fileSetIds
        clArgPiece = clPathPiece + clArgPiece

    if '' != args.png_path:
        clPathPiece = " -p pngPath=" + args.png_path
        # insert the casa parameter here, since it is fixed for all fileSetIds
        clArgPiece = clPathPiece + clArgPiece


    clFinish = " runRealFastIngestionWorkflow"


    #
    # Loop over the provided fileSetIds and send a command for each:
    #
    for id in args.sdmName:
        wfArgs = clBase + os.environ['CAPO_PROFILE'] + clArgPiece + id + clFinish
        print("wf "+wfArgs)
        wf(args=wfArgs.split(' '))
        time.sleep(0.25) # 1/4 second delay between workflow launches for safety
