""" Test case for an ingest path with files like those in
/lustre/aoc/cluster/pipeline/dsoc-dev/workspaces/staging/tmpx_ratuqh"""

from pathlib import Path

import json

import shutil

# pylint: disable=E0401, E0402, W0621

import pytest

from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.manifest_components import (
    TARFILE_EXT,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, Telescope, ScienceProductType
from .conftest import (
    ingest_path,
    ADDITIONAL_METADATA_FILENAME,
    OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES,
    ANCILLARY_PRODUCTS,
    populate_fake_tmpx_ratuqh_ingest_path,
    OUTPUT_SCIENCE_PRODUCT,
)


def test_ancillary_products_rendered_correctly(ingest_path: Path):
    """
    The ancillary_products section of the manifest we build
    should match the one in the example.

    :param ingest_path: staging source dir
    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)

    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)

    output_group = manifest.output_group
    aps = output_group.ancillary_products
    assert len(aps) == len(ANCILLARY_PRODUCTS)
    ap_data = []

    for ap in aps:
        ap_data.append({"filename": ap.filename, "json": ap.to_json()})
    assert len(ap_data) == len(aps)

    # make sure all the ancillary products were created...
    filename_count = 0
    for product in ANCILLARY_PRODUCTS:
        # (It won't find the ingest artifact tar yet, because we didn't populate
        # the ingest path with it; it's produced during manifest creation.)

        # ...and that each one's JSON is well formed.
        for properties in ap_data:
            if properties["filename"] == product.filename:
                filename_count += 1
                # If there's a class `repr` in there rather than JSON-serializable text,
                # dump will fail
                json.dumps(properties["json"])

    ingest_artifacts_tar = find_ingestion_artifacts_tar(ingest_path)
    assert ingest_artifacts_tar
    filename_count += 1

    assert filename_count == len(ANCILLARY_PRODUCTS)

    shutil.rmtree(ingest_path)


@pytest.mark.skip("TODO NEXT")
def test_output_science_products_rendered_correctly(ingest_path: Path):
    """
    The output_group section of the manifest we build
    should match the one in the example.

    :param ingest_path: staging source dir
    :return:
    """
    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    manifest, _ = build_tmpx_ratuqh_image_manifest(ingest_path)
    output_group = manifest.output_group

    science_products = output_group.science_products
    assert len(science_products) == 1
    science_product = science_products[0]
    # TODO: where the hell is that quicklook image coming from??
    assert science_product == OUTPUT_SCIENCE_PRODUCT

    # TODO: why does science product have no ancillaries?
    sp_ancillaries = science_product.ancillary_products
    assert len(sp_ancillaries) == len(OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES)
    assert sp_ancillaries == OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES

    # TODO: where the hell are the ancillary products in this thing?
    other_ancillaries = output_group.ancillary_products
    assert len(other_ancillaries) == len(ANCILLARY_PRODUCTS)
    assert other_ancillaries == ANCILLARY_PRODUCTS

    shutil.rmtree(ingest_path)


@pytest.mark.skip("TODO")
def test_ing_artifacts_tar_has_only_what_it_should():
    """
    The ingestion_artifacts tar should contain ONLY
    the ingestion manifest and the additional metadata file

    :return:
    """
    # TODO:

    shutil.rmtree(ingest_path)
    raise NotImplementedError


# -----------------------------
#  U  T  I  L  I  T  I  E  S
# -----------------------------


def build_tmpx_ratuqh_image_manifest(staging_source_dir: Path):
    """
    Build an image ingestion manifest that "should" match image_manifest_tmpx_ratuqh.json

    :return: the ingestion manifest and the manifest file in which it lives
    """

    additional_metadata = AbstractTextFile(filename=ADDITIONAL_METADATA_FILENAME, content="")

    return IngestionManifestBuilder(
        staging_source_dir=staging_source_dir,
        additional_metadata=additional_metadata,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="uid://evla/calibration/ea93dae5-3495-47fa-887d-4be2852f5f14",
    ).build()


def find_ingestion_artifacts_tar(staging_source_dir: Path):
    """
    There should be an ingestion artifacts tar after manifest creation.

    :param staging_source_dir:
    :return:
    """
    ing_artifacts_tars = [
        file
        for file in staging_source_dir.iterdir()
        if file.name.startswith(AncillaryProductType.INGESTION_ARTIFACTS.value)
        and file.name.endswith(TARFILE_EXT)
    ]
    assert len(ing_artifacts_tars) == 1
    return ing_artifacts_tars[0]
